#!/usr/bin/env python

__author__ = "Patrick Godwin (patrick.godwin@ligo.org)"
__description__ = "a module for influxdb I/O utilities"

#-------------------------------------------------
### imports

from collections import defaultdict
import itertools
import json

from six.moves import http_client
from six.moves import urllib

import numpy

import lal

from .. import aggregator
from . import line_protocol
from .. import utils


#-------------------------------------------------
### classes

class InfluxDBAggregator(object):
    """
    Handles the storing and aggregation of timeseries into InfluxDB.
    """
    def __init__(self, hostname='localhost', port=8086, db='mydb', reduce_dt=300, reduce_across_tags=True, **kwargs):
        self.hostname = hostname
        self.port = port
        self.db = db

        ### reduction options
        self.reduce_dt = reduce_dt
        self.reduce_across_tags = reduce_across_tags

        ### set up client and database
        self.client = create_client(host=self.hostname, port=self.port)
        set_up_database(self.client, self.db)

        ### track reduced data to process
        self.span_processed = defaultdict(list)
        self.last_reduce = aggregator.now()


    def store_and_reduce(self, measurement, data, columns, tags=None, aggregates=('min', 'median', 'max')):
        """
        Stores and aggregates incoming timeseries into InfluxDB.

        Parameters
        ----------
        measurement : `str`
            the measurement name
        data : {`tuple` or `str` : `list`}
            see below for guide/formatting
        columns : iterable or 'str'
            a list of column names (not including time) to store
        tags : `tuple` or `str`
            tag keys corresponding to tag values in data
        aggregates : iterable or `str`
            aggregate quantities to use in downsampling (i.e. max)

        Given N columns and M tags, data is passed in as follows:
            {(tag1, ..., tagM): [iterable(time), iterable(column1), ..., iterable(columnN)]}

        where each iterable is of the same length.

        In the case where no tags are passed in, data is instead passed in as a list:
            [iterable(time), iterable(column1), ..., iterable(columnN)]

        """
        if isinstance(columns, str):
            columns = [columns]
        if tags and isinstance(tags, str):
            tags = [tags]
        if isinstance(aggregates, str):
            aggregates = [aggregates]

        this_span = []
        for tag_key, data_entry in data.items():
            time = data_entry[0]
            if len(time) > 0:
                start, end = numpy.floor(time[0]).astype(int), numpy.ceil(time[-1]).astype(int)

                if not isinstance(tag_key, tuple):
                    tag_key = (tag_key,)
                tag_entry = {tag_name: tag_val for tag_name, tag_val in zip(tags, tag_key)}

                ### reduce to 1s by default before storing timeseries
                for aggregate in aggregates:
                    aggfunc = aggregator.aggregate_to_func(aggregate)

                    for ii, column in enumerate(columns, 1):
                        ### store and reduce
                        reduced_time, reduced_data = aggregator.reduce_data(time, data_entry[ii], aggfunc, dt=1)
                        store_timeseries(self.client, self.db, measurement, reduced_time, {column: reduced_data}, tags=tag_entry, aggregate=aggregate, dt=1)

                        ### keep track of spans processed
                        if self.span_processed[(measurement, column, tags[0])]:
                            prev_start, prev_end = self.span_processed[(measurement, column, tags[0])]
                            self.span_processed[(measurement, column, tags[0])] = [min(start, prev_start), max(end, prev_end)]
                        else:
                            self.span_processed[(measurement, column, tags[0])] = [start, end]

                        if this_span:
                            this_span = [min(start, this_span[0]), max(end, this_span[1])]
                        else:
                            this_span = [start, end]

        ### reduce timeseries
        if this_span:
            for column in columns:

                ### reduce across tag for dt=1
                if self.reduce_across_tags:
                    start, end = utils.span_to_process(*this_span, dt=1)
                    #print('processing column: %s, dt: %d, spans: %d - %d'%(column, 1, start, end))
                    for aggregate in aggregates:
                        reduce_across_tags(self.client, self.db, measurement, start, end, column, tags[0], aggregate, 1)

                ### only reduce higher dt every last_reduce seconds
                if utils.in_new_epoch(aggregator.now(), self.last_reduce, self.reduce_dt):
                    self.last_reduce = aggregator.now()

                    ### reduce by tag
                    for dt in (10**power for power in range(aggregator.DIRS - 1)):
                        start, end = utils.span_to_process(*self.span_processed[(measurement, column, tags[0])], dt=10*dt)
                        #print('processing column: %s, dt: %d, spans: %d - %d'%(column, 10*dt, start, end))
                        for aggregate in aggregates:
                            reduce_by_tag(self.client, self.db, measurement, start, end, column, tags[0], aggregate, dt)

                    ### reduce across tags
                    if self.reduce_across_tags:
                        for dt in (10**power for power in range(1, aggregator.DIRS)):

                            start, end = utils.span_to_process(*self.span_processed[(measurement, column, tags[0])], dt=dt)
                            #print('processing column: %s, dt: %d, spans: %d - %d'%(column, dt, start, end))
                            for aggregate in aggregates:
                                reduce_across_tags(self.client, self.db, measurement, start, end, column, tags[0], aggregate, dt)

                    ### reset processed spans
                    self.span_processed[(measurement, column, tags[0])] = []

        this_span = []


#-------------------------------------------------
### i/o utilities

def store_timeseries(client, db, measurement, time, data, tags=None, aggregate=None, dt=None):
    """Insert timeseries into InfluxDB.

    Parameters
    ----------
    client : `HTTPConnection`
        a long-lived HTTP connection to an InfluxDB database
    db : `str`
        the database name to store timeseries into
    measurement : `str`
        the measurement name
    time : `list`
        GPS start time
    data : {`str` : `list`}
        key-value pairs in the form: column, (list of values)
    tags : {`str` :  `list`}
        key-value pairs in the form: tag_name, tag_value
    aggregate : `str`
        aggregate quantities to use in downsampling (i.e. max)
    dt : `int`
        the retention policy in which to store timeseries in,
        not used if aggregate is not specified

    """

    ### format tags
    if aggregate:
        tag_dict = {'aggregate': aggregate}
    else:
        tag_dict = {}
    if tags:
        tag_dict.update(tags)

    ### convert to unix time
    time = [utils.gps_to_unix(t) for t in time]

    ### format timeseries
    points = []
    rows = [dict(zip(data, row)) for row in zip(*data.values())]
    for t, row in zip(time, rows):
        point = {
            "measurement": measurement,
            "tags": tag_dict,
            "time": t,
            "fields": row,
        }
        points.append(point)

    packet = {'points':points, 'tags':tags}

    ### convert to line protocol
    lines = line_protocol.make_lines(packet).encode('utf-8')

    ### push to client
    param_dict = {'db': db}

    if dt:
        param_dict.update({'rp': '{}s'.format(dt)})

    params = urllib.parse.urlencode(param_dict)
    client.request('POST', '/write?{}'.format(params), lines, headers={"Content-Type": "application/octet-stream"})
    response = client.getresponse().read()


def retrieve_binnedtimeseries_by_tag(client, db, measurement, start, end, column, tag, aggregate=None, dt=None, datetime=False):
    """Retrieve all timeseries with a given tag from InfluxDB, binned by tag and dt.

    Parameters
    ----------
    client : `HTTPConnection`
        a long-lived HTTP connection to an InfluxDB database
    db : `str`
        the database name to store timeseries into
    measurement : `str`
        the measurement name
    start : `int`
        GPS start time
    end : `int`
        GPS end time
    tag : `str`:
        tag to match timeseries with
    aggregate : `str`
        aggregate quantities to use in downsampling (i.e. max)
    dt : `int`
        the retention policy in which to store timeseries in,
        not used if aggregate is not specified

    Returns
    -------
    timebins : `list`
        time bins spanning from start to end (inclusive) with spacing dt
    tag_values : `list`
        all tag values that matched tag
    binned_data: `list` of `list`
        a list of timeseries, each one corresponding to a tag value ordered
        by tag_values

    """
    timedata, tagdata, datadata = retrieve_timeseries_by_tag(
        client,
        db,
        measurement,
        start,
        end,
        column,
        tag,
        aggregate=aggregate,
        dt=dt,
        datetime=False,
    )

    ### create time bins
    if dt:
        binned_timedata = (numpy.arange((end-start)/dt+1)*dt + start).astype(int)
    else:
        binned_timedata = (numpy.arange(end-start) + start).astype(int)

    binned_datadata = numpy.zeros((len(tagdata), len(binned_timedata)), dtype="double") + float("nan")

    ### bin time and data
    for i, (time, tag, data) in enumerate(zip(timedata, tagdata, datadata)):
        if len(time) > 0:
            ix = numpy.digitize(time, binned_timedata) - 1
            binned_datadata[i,ix] = data

    ### convert to datetime (if requested)
    ### FIXME: need to create a gps_to_rfc3339 function
    if datetime:
        raise NotImplementedError('gps_to_rfc3339 functionality to do this is not currently implemented')
    #    binned_timedata = [utils.gps_to_rfc3339(t) for t in binned_timedata]

    return binned_timedata.tolist(), tagdata, binned_datadata.tolist()


def retrieve_timeseries_by_tag(client, db, measurement, start, end, column, tag, aggregate=None, dt=None, datetime=False):
    """Retrieve all timeseries with a given tag from InfluxDB.

    Parameters
    ----------
    client : `HTTPConnection`
        a long-lived HTTP connection to an InfluxDB database
    db : `str`
        the database name to store timeseries into
    measurement : `str`
        the measurement name
    start : `int`
        GPS start time
    end : `int`
        GPS end time
    tag : `str`:
        tag to match timeseries with
    aggregate : `str`
        aggregate quantities to use in downsampling (i.e. max)
    dt : `int`
        the retention policy in which to store timeseries in,
        not used if aggregate is not specified

    Returns
    -------
    time : `list`
        time points spanning from start to end
    tag_values : `list`
        all tag values that matched tag
    data: `list` of `list`
        a list of timeseries, each one corresponding to a tag value ordered
        by tag_values

    """
    if aggregate:
        agg_condition = 'aggregate = \'{aggregate}\' AND'.format(aggregate=aggregate)
    else:
        agg_condition = ''

    if dt:
        retention_policy = '"{}s"'.format(dt)
    else:
        retention_policy = ''

    ### convert to unix time
    start = utils.gps_to_unix(start)
    end = utils.gps_to_unix(end)

    query = ' SELECT "{column}","{tag}" FROM "{db}".{retention_policy}."{measurement}" WHERE {agg_condition} time >= {start} AND time <= {end}'.format(
        column=column,
        tag=tag,
        db=db,
        retention_policy=retention_policy,
        measurement=measurement,
        agg_condition=agg_condition,
        start=start,
        end=end,
    )

    if datetime:
        epoch = 'rfc3339'
    else:
        epoch = 'ns'

    params = urllib.parse.urlencode({'db':db, 'q':query, 'epoch':epoch})
    client.request('GET', '/query?{}'.format(params), headers={"Content-Type": "application/json"})
    response = json.loads(client.getresponse().read().decode('utf-8'))
    try:
        points = response['results'][0]['series'][0]['values']
    except:
        return [], [], []
    else:
        tags = set()
        series = defaultdict(lambda: defaultdict(list))
        for time, data, tag_val in points:
            series[tag_val]['time'].append(time)
            series[tag_val][column].append(data)
            tags.add(tag_val)

        tagdata = sorted(list(tags))
        timedata = []
        datadata = []
        for tag_val in tagdata:
            timedata.append(series[tag_val]['time'])
            datadata.append(series[tag_val][column])

        ### convert to gps time
        if not datetime:
            timedata = [utils.unix_to_gps(numpy.array(time)).tolist() for time in timedata]

        return timedata, tagdata, datadata


def retrieve_timeseries_latest(client, db, measurement, column, tags=None, aggregate=None, dt=None, datetime=False, num_latest=1):
    if aggregate:
        agg_condition = 'aggregate = \'{aggregate}\''.format(aggregate=aggregate)
    else:
        agg_condition = ''
    if tags:
        tag_filters = ''.join([' AND {} = \'{}\''.format(tag, val) for tag, val in tags])
    else:
        tag_filters = ''

    if dt:
        retention_policy = '"{}s"'.format(dt)
    else:
        retention_policy = ''

    if datetime:
        epoch = 'rfc3339'
    else:
        epoch = 'ns'

    query = ' SELECT "{column}" FROM "{db}".{retention_policy}."{measurement}" WHERE {agg_condition} {tag_filters} ORDER BY time DESC LIMIT {num_latest}'.format(
        column=column,
        db=db,
        retention_policy=retention_policy,
        measurement=measurement,
        agg_condition=agg_condition,
        tag_filters=tag_filters,
        num_latest=num_latest,
    )

    params = urllib.parse.urlencode({'db':db, 'q':query, 'epoch':epoch})
    client.request('GET', '/query?{}'.format(params), headers={"Content-Type": "application/json"})
    response = json.loads(client.getresponse().read().decode('utf-8'))

    try:
        points = response['results'][0]['series'][0]['values']
    except:
        return [], []
    else:
        time = []
        data = []
        for t, d in points:
            time.append(t)
            data.append(d)

        ### convert to gps time
        if not datetime:
            time = utils.unix_to_gps(numpy.array(time)).tolist()

        return time, data


def retrieve_timeseries(client, db, measurement, start, end, column, tags=None, aggregate=None, dt=None, datetime=False):
    if aggregate:
        agg_condition = 'aggregate = \'{aggregate}\' AND'.format(aggregate=aggregate)
    else:
        agg_condition = ''
    if tags:
        tag_filters = ''.join(['{} = \'{}\' AND '.format(tag, val) for tag, val in tags])
    else:
        tag_filters = ''

    if dt:
        retention_policy = '"{}s"'.format(dt)
    else:
        retention_policy = ''

    if datetime:
        epoch = 'rfc3339'
    else:
        epoch = 'ns'

    ### convert to unix time
    start = utils.gps_to_unix(start)
    end = utils.gps_to_unix(end)

    query = ' SELECT "{column}" FROM "{db}".{retention_policy}."{measurement}" WHERE {agg_condition} {tag_filters} time >= {start} AND time <= {end}'.format(
        column=column,
        db=db,
        retention_policy=retention_policy,
        measurement=measurement,
        agg_condition=agg_condition,
        tag_filters=tag_filters,
        start=start,
        end=end,
    )

    params = urllib.parse.urlencode({'db':db, 'q':query, 'epoch':epoch})
    client.request('GET', '/query?{}'.format(params), headers={"Content-Type": "application/json"})
    response = json.loads(client.getresponse().read().decode('utf-8'))

    try:
        points = response['results'][0]['series'][0]['values']
    except:
        return [], []
    else:
        time = []
        data = []
        for t, d in points:
            time.append(t)
            data.append(d)

        ### convert to gps time
        if not datetime:
            time = utils.unix_to_gps(numpy.array(time)).tolist()

        return time, data


#-------------------------------------------------
### reduction utilities

def reduce_by_tag(client, db, measurement, start, end, column, tag, aggregate, dt):
    timedata, tagdata, datadata = retrieve_timeseries_by_tag(client, db, measurement, start, end, column, tag, aggregate=aggregate, dt=dt)

    reduced_dt = 10 * dt
    aggfunc = aggregator.aggregate_to_func(aggregate)

    ### reduce by tag
    for tag_name, time, data in zip(tagdata, timedata, datadata):
        reduced_time, reduced_data = aggregator.reduce_data(time, data, aggfunc, reduced_dt)
        store_timeseries(client, db, measurement, reduced_time, {column: reduced_data}, tags={tag: tag_name}, aggregate=aggregate, dt=reduced_dt)


def reduce_across_tags(client, db, measurement, start, end, column, tag, aggregate, dt):
    timedata, tagdata, datadata = retrieve_timeseries_by_tag(client, db, measurement, start, end, column, tag, aggregate=aggregate, dt=dt)

    aggfunc = aggregator.aggregate_to_func(aggregate)

    ### flatten data and prep to reduce
    time = list(itertools.chain.from_iterable(timedata))
    data = list(itertools.chain.from_iterable(datadata))
    agg_measurement = '{}_across_{}s'.format(measurement, tag)

    ### reduce across tags
    reduced_time, reduced_data = aggregator.reduce_data(time, data, aggfunc, dt=dt)
    store_timeseries(client, db, agg_measurement, reduced_time, {column: reduced_data}, aggregate=aggregate, dt=dt)


#-------------------------------------------------
### database utilities

def set_up_database(client, db):
    create_database(client, db)
    dts = [10**power for power in range(aggregator.DIRS)]
    for dt in dts:
        create_retention_policy(client, db, dt)


def create_retention_policy(client, db, dt):
    query = ' CREATE RETENTION POLICY "{dt}s" ON "{db}" DURATION INF REPLICATION 1'.format(db=db, dt=dt)

    params = urllib.parse.urlencode({'q':query})
    client.request('POST', '/query?{}'.format(params), headers={"Content-Type": "application/json"})
    response = client.getresponse().read()


def create_database(client, db):
    query = ' CREATE DATABASE "{db}"'.format(db=db)

    params = urllib.parse.urlencode({'q':query})
    client.request('POST', '/query?{}'.format(params), headers={"Content-Type": "application/json"})
    response = client.getresponse().read()


def create_client(host='localhost', port=8086):
    return http_client.HTTPConnection(host, port)
